@@ -3,7 +3,7 @@
 //! Gibbs Sea Water Oceanographic Toolbox of TEOS-10 implemented in Rust.
 //! version: 3.06.12
 //!
-//! http://www.teos-10.org
+//! <http://www.teos-10.org>
 //!
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -15,6 +15,10 @@
 #[allow(unused)]
 mod gsw_internal_const;
 
+/// cbindgen:ignore
+#[allow(unused)]
+mod gsw_sp_coefficients;
+
 /// cbindgen:ignore
 #[allow(unused)]
 mod gsw_specvol_coefficients;
@@ -25,6 +29,7 @@ mod ffi;
 pub mod conversions;
 pub mod earth;
 mod gsw_internal_funcs;
+pub mod practical_salinity;
 pub mod volume;
 
 mod error;
