@@ -26,8 +26,35 @@ fn non_dimensional_sa(sa: f64) -> Result<f64> {
     Ok(libm::sqrt(GSW_SFAC * sa + OFFSET))
 }
 
+/// in-situ density
+///
+/// Calculates in-situ density from Absolute Salinity and Conservative
+/// Temperature, using the computationally-efficient expression for
+/// specific volume in terms of SA, CT and p (Roquet et al., 2014).
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+///
+/// # Returns
+///
+/// * `rho`: in-situ density \[kg m-3\]
+///
+pub fn rho(sa: f64, ct: f64, p: f64) -> Result<f64> {
+    Ok(1.0 / specvol(sa, ct, p)?)
+}
+
 /// Thermal expansion coefficient with respect to Conservative Temperature
 /// (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[ g kg-1 \]
+/// * `ct`: Conservative Temperature (ITS-90) \[ deg C \]
+/// * `p`: sea pressure \[ dbar \] (i.e. absolute pressure - 10.1325 dbar)
+///
 pub fn alpha(sa: f64, ct: f64, p: f64) -> Result<f64> {
     let xs: f64 = non_dimensional_sa(sa)?;
     let ys: f64 = ct / GSW_CTU;
@@ -55,6 +82,13 @@ pub fn alpha(sa: f64, ct: f64, p: f64) -> Result<f64> {
 
 /// Saline contraction coefficient at constant Conservative Temperature
 /// (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[ g kg-1 \]
+/// * `ct`: Conservative Temperature (ITS-90) \[ deg C \]
+/// * `p`: sea pressure \[ dbar \] (i.e. absolute pressure - 10.1325 dbar)
+///
 pub fn beta(sa: f64, ct: f64, p: f64) -> Result<f64> {
     let xs: f64 = non_dimensional_sa(sa)?;
     let ys: f64 = ct / GSW_CTU;
@@ -80,6 +114,38 @@ pub fn beta(sa: f64, ct: f64, p: f64) -> Result<f64> {
     Ok(-v_sa * 0.5 * GSW_SFAC / (specvol(sa, ct, p)? * xs))
 }
 
+/// in-situ density, thermal expansion & saline contraction coefficients
+/// (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[ g kg-1 \]
+/// * `ct`: Conservative Temperature (ITS-90) \[ deg C \]
+/// * `p`: sea pressure \[ dbar \] (i.e. absolute pressure - 10.1325 dbar)
+///
+pub fn rho_alpha_beta(sa: f64, ct: f64, p: f64) -> Result<(f64, f64, f64)> {
+    let rho = rho(sa, ct, p)?;
+    let alpha = alpha(sa, ct, p)?;
+    let beta = beta(sa, ct, p)?;
+
+    Ok((rho, alpha, beta))
+}
+
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[ g kg-1 \]
+/// * `ct`: Conservative Temperature (ITS-90) \[ deg C \]
+/// * `p`: sea pressure \[ dbar \] (i.e. absolute pressure - 10.1325 dbar)
+///
+pub fn specvol_alpha_beta(sa: f64, ct: f64, p: f64) -> Result<(f64, f64, f64)> {
+    let specvol = specvol(sa, ct, p)?;
+    let alpha = alpha(sa, ct, p)?;
+    let beta = beta(sa, ct, p)?;
+
+    Ok((specvol, alpha, beta))
+}
+
 /// Specific volume of sea water (75-term polynomial approximation)
 ///
 /// Calculates specific volume from Absolute Salinity, Conservative
@@ -138,6 +204,67 @@ pub fn specvol(sa: f64, ct: f64, p: f64) -> Result<f64> {
                     + z * (V004 + xs * V104 + ys * V014 + z * (V005 + z * V006))))))
 }
 
+#[cfg(test)]
+mod test_specvol {
+    use super::specvol;
+
+    #[test]
+    // NaN input results in NaN output.
+    // Other libraries using GSW-rs might rely on this behavior to propagate
+    // and handle invalid elements.
+    fn nan() {
+        let v = specvol(f64::NAN, 1.0, 1.0);
+        assert!(v.unwrap().is_nan());
+
+        let v = specvol(1.0, f64::NAN, 1.0);
+        assert!(v.unwrap().is_nan());
+
+        let v = specvol(1.0, 1.0, f64::NAN);
+        assert!(v.unwrap().is_nan());
+    }
+
+    #[test]
+    // Test value from Roquet 2015 Appendix C.3, rounded to 9.732819628e-04
+    fn roquet2015() {
+        assert!((specvol(30., 10., 1000.0).unwrap() - 9.732819628e-04).abs() <= 5e-14);
+    }
+
+    #[allow(clippy::excessive_precision)]
+    #[test]
+    fn test_specvol() {
+        if cfg!(feature = "compat") {
+            // Test value from C library.
+            assert!(
+                (specvol(34.507499465692057, 27.994827331978655, 0.0).unwrap()
+                    - 0.00097855432330275953)
+                    .abs()
+                    < f64::EPSILON
+            );
+        }
+    }
+}
+
+/// Specific volume anomaly (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[ g kg-1 \]
+/// * `ct`: Conservative Temperature (ITS-90) \[ deg C \]
+/// * `p`: sea pressure \[ dbar \] (i.e. absolute pressure - 10.1325 dbar)
+/// * `sa_ref`:
+/// * `ct_ref`:
+///
+fn specvol_anom(sa: f64, ct: f64, p: f64, sa_ref: Option<f64>, ct_ref: Option<f64>) -> Result<f64> {
+    let xs: f64 = non_dimensional_sa(sa)?;
+    let ys: f64 = ct / GSW_CTU;
+    let z: f64 = non_dimensional_p(p);
+
+    unimplemented!();
+
+    // let xs_ref: f64 = non_dimensional_sa(sa_ref)?;
+    // let ys: f64 = ct_ref / GSW_CTU;
+}
+
 /// Specific Volume Anomaly of Standard Ocean Salinity and CT=0
 ///
 /// Specific volume anomaly with reference of SA = SSO & CT = 0 (75-term equation)
@@ -156,128 +283,80 @@ pub fn specvol_anom_standard(sa: f64, ct: f64, p: f64) -> Result<f64> {
     Ok(specvol(sa, ct, p)? - crate::gsw_internal_funcs::specvol_sso_0(p))
 }
 
-pub fn specvol_first_derivatives(sa: f64, ct: f64, p: f64) -> Result<(f64, f64, f64)> {
-    let xs: f64 = non_dimensional_sa(sa)?;
-    let ys: f64 = ct / GSW_CTU;
-    let z: f64 = non_dimensional_p(p);
-
-    let v_ct_part: f64 = A000
-        + xs * (A100 + xs * (A200 + xs * (A300 + xs * (A400 + A500 * xs))))
-        + ys * (A010
-            + xs * (A110 + xs * (A210 + xs * (A310 + A410 * xs)))
-            + ys * (A020
-                + xs * (A120 + xs * (A220 + A320 * xs))
-                + ys * (A030 + xs * (A130 + A230 * xs) + ys * (A040 + A140 * xs + A050 * ys))))
-        + z * (A001
-            + xs * (A101 + xs * (A201 + xs * (A301 + A401 * xs)))
-            + ys * (A011
-                + xs * (A111 + xs * (A211 + A311 * xs))
-                + ys * (A021 + xs * (A121 + A221 * xs) + ys * (A031 + A131 * xs + A041 * ys)))
-            + z * (A002
-                + xs * (A102 + xs * (A202 + A302 * xs))
-                + ys * (A012 + xs * (A112 + A212 * xs) + ys * (A022 + A122 * xs + A032 * ys))
-                + z * (A003 + A103 * xs + A013 * ys + A004 * z)));
-
-    let v_ct = 0.025 * v_ct_part;
-
-    let v_sa_part: f64 = B000
-        + xs * (B100 + xs * (B200 + xs * (B300 + xs * (B400 + B500 * xs))))
-        + ys * (B010
-            + xs * (B110 + xs * (B210 + xs * (B310 + B410 * xs)))
-            + ys * (B020
-                + xs * (B120 + xs * (B220 + B320 * xs))
-                + ys * (B030 + xs * (B130 + B230 * xs) + ys * (B040 + B140 * xs + B050 * ys))))
-        + z * (B001
-            + xs * (B101 + xs * (B201 + xs * (B301 + B401 * xs)))
-            + ys * (B011
-                + xs * (B111 + xs * (B211 + B311 * xs))
-                + ys * (B021 + xs * (B121 + B221 * xs) + ys * (B031 + B131 * xs + B041 * ys)))
-            + z * (B002
-                + xs * (B102 + xs * (B202 + B302 * xs))
-                + ys * (B012 + xs * (B112 + B212 * xs) + ys * (B022 + B122 * xs + B032 * ys))
-                + z * (B003 + B103 * xs + B013 * ys + B004 * z)));
-
-    let v_sa = 0.5 * GSW_SFAC * v_sa_part / xs;
-
-    let v_p_part = C000
-        + xs * (C100 + xs * (C200 + xs * (C300 + xs * (C400 + C500 * xs))))
-        + ys * (C010
-            + xs * (C110 + xs * (C210 + xs * (C310 + C410 * xs)))
-            + ys * (C020
-                + xs * (C120 + xs * (C220 + C320 * xs))
-                + ys * (C030 + xs * (C130 + C230 * xs) + ys * (C040 + C140 * xs + C050 * ys))))
-        + z * (C001
-            + xs * (C101 + xs * (C201 + xs * (C301 + C401 * xs)))
-            + ys * (C011
-                + xs * (C111 + xs * (C211 + C311 * xs))
-                + ys * (C021 + xs * (C121 + C221 * xs) + ys * (C031 + C131 * xs + C041 * ys)))
-            + z * (C002
-                + xs * (C102 + C202 * xs)
-                + ys * (C012 + C112 * xs + C022 * ys)
-                + z * (C003 + C103 * xs + C013 * ys + z * (C004 + C005 * z))));
-
-    let v_p = 1e-8 * v_p_part;
-
-    Ok((v_sa, v_ct, v_p))
-}
-
-pub fn specvol_alpha_beta(sa: f64, ct: f64, p: f64) -> Result<(f64, f64, f64)> {
-    let specvol = specvol(sa, ct, p)?;
-    let alpha = alpha(sa, ct, p)?;
-    let beta = beta(sa, ct, p)?;
-
-    Ok((specvol, alpha, beta))
-}
-
-/*
-fn specvol_first_derivatives_wrt_enthalpy(sa: f64, ct: f64, p: f64) -> (f64, f64) {
-    unimplemented!()
-}
-
-fn specvol_ice(t: f64, p: f64) {
-    unimplemented!()
-}
-
-fn specvol_second_derivatives(sa: f64, ct: f64, p: f64) -> (f64, f64, f64, f64, f64) {
-    unimplemented!()
-}
-
-fn specvol_second_derivatives_wrt_enthalpy(sa: f64, ct: f64, p: f64) -> (f64, f64, f64) {
-    unimplemented!()
-}
-
-fn specvol_diff() {
-    unimplemented!()
-}
-fn specvol_from_pot_enthalpy_ice() {
-    unimplemented!()
-}
-fn specvol_from_pot_enthalpy_ice_poly() {
-    unimplemented!()
-}
-fn specvol_p_parts() {
-    unimplemented!()
-}
-*/
-
-/// in-situ density
-///
-/// Calculates in-situ density from Absolute Salinity and Conservative
-/// Temperature, using the computationally-efficient expression for
-/// specific volume in terms of SA, CT and p (Roquet et al., 2014).
+/// Potential density anomaly with reference to sea pressure of 0 dbar
+/// (75-term polynomial approximation)
 ///
 /// # Arguments
 ///
 /// * `sa`: Absolute Salinity \[g kg-1\]
 /// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
 /// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+pub fn sigma0(sa: f64, ct: f64) -> Result<f64> {
+    let xs: f64 = non_dimensional_sa(sa)?;
+    let ys: f64 = ct / GSW_CTU;
+
+    // Specific Volume
+    let v = V000
+        + xs * (V100 + xs * (V200 + xs * (V300 + xs * (V400 + xs * (V500 + xs * V600)))))
+        + ys * (V010
+            + xs * (V110 + xs * (V210 + xs * (V310 + xs * (V410 + xs * V510))))
+            + ys * (V020
+                + xs * (V120 + xs * (V220 + xs * (V320 + xs * V420)))
+                + ys * (V030
+                    + xs * (V130 + xs * (V230 + xs * V330))
+                    + ys * (V040
+                        + xs * (V140 + xs * V240)
+                        + ys * (V050 + xs * V150 + ys * V060)))));
+
+    Ok(1.0 / v - 1000.0)
+}
+
+/// Potential density anomaly with reference to sea pressure of 1000 dbar
+/// (75-term polynomial approximation)
 ///
-/// # Returns
+/// # Arguments
 ///
-/// * `rho`: in-situ density \[kg m-3\]
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+pub fn sigma1(sa: f64, ct: f64) -> Result<f64> {
+    Ok(rho(sa, ct, 1000.0)? - 1000.0)
+}
+
+/// Potential density anomaly with reference to sea pressure of 2000 dbar
+/// (75-term polynomial approximation)
 ///
-pub fn rho(sa: f64, ct: f64, p: f64) -> Result<f64> {
-    Ok(1.0 / specvol(sa, ct, p)?)
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+pub fn sigma2(sa: f64, ct: f64) -> Result<f64> {
+    Ok(rho(sa, ct, 2000.0)? - 1000.0)
+}
+
+/// Potential density anomaly with reference to sea pressure of 3000 dbar
+/// (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+pub fn sigma3(sa: f64, ct: f64) -> Result<f64> {
+    Ok(rho(sa, ct, 3000.0)? - 1000.0)
+}
+
+/// Potential density anomaly with reference to sea pressure of 4000 dbar
+/// (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+pub fn sigma4(sa: f64, ct: f64) -> Result<f64> {
+    Ok(rho(sa, ct, 4000.0)? - 1000.0)
 }
 
 /// Sound speed in seawater (75-term polynomial approximation)
@@ -301,7 +380,7 @@ pub fn rho(sa: f64, ct: f64, p: f64) -> Result<f64> {
 /// * IOC, SCOR and IAPSO, 2010: The international thermodynamic equation of
 ///   seawater - 2010: Calculation and use of thermodynamic properties.
 ///   Intergovernmental Oceanographic Commission, Manuals and Guides No. 56,
-///   UNESCO (English), 196 pp.  Available from http://www.TEOS-10.org
+///   UNESCO (English), 196 pp.  Available from <http://www.TEOS-10.org>
 ///   See Appendix K of this TEOS-10 Manual.
 ///
 /// * McDougall, T.J., D.R. Jackett, D.G. Wright and R. Feistel, 2003:
@@ -372,77 +451,210 @@ pub fn sound_speed(sa: f64, ct: f64, p: f64) -> Result<f64> {
     Ok(10_000.0 * libm::sqrt(-v * v / v_p))
 }
 
-/// Potential density anomaly with reference to sea pressure of 0 dbar
-/// (75-term polynomial approximation)
-pub fn sigma0(sa: f64, ct: f64) -> Result<f64> {
-    let xs: f64 = non_dimensional_sa(sa)?;
-    let ys: f64 = ct / GSW_CTU;
-
-    // Specific Volume
-    let v = V000
-        + xs * (V100 + xs * (V200 + xs * (V300 + xs * (V400 + xs * (V500 + xs * V600)))))
-        + ys * (V010
-            + xs * (V110 + xs * (V210 + xs * (V310 + xs * (V410 + xs * V510))))
-            + ys * (V020
-                + xs * (V120 + xs * (V220 + xs * (V320 + xs * V420)))
-                + ys * (V030
-                    + xs * (V130 + xs * (V230 + xs * V330))
-                    + ys * (V040
-                        + xs * (V140 + xs * V240)
-                        + ys * (V050 + xs * V150 + ys * V060)))));
-
-    Ok(1.0 / v - 1000.0)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+///
+fn internal_energy(sa: f64, ct: f64, p: f64) -> Result<f64> {
+    unimplemented!()
 }
 
-/// Potential density anomaly with reference to sea pressure of 1000 dbar
-/// (75-term polynomial approximation)
-pub fn sigma1(sa: f64, ct: f64) -> Result<f64> {
-    Ok(rho(sa, ct, 1000.0)? - 1000.0)
+/// Specific enthalpy of seawater (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+///
+fn enthalpy(sa: f64, ct: f64, p: f64) -> Result<f64> {
+    Ok(GSW_CP0 * ct + dynamic_enthalpy(sa, ct, p)?)
 }
 
-/// Potential density anomaly with reference to sea pressure of 2000 dbar
-/// (75-term polynomial approximation)
-pub fn sigma2(sa: f64, ct: f64) -> Result<f64> {
-    Ok(rho(sa, ct, 2000.0)? - 1000.0)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+///
+fn enthalpy_diff(sa: f64, ct: f64, p: f64) -> Result<f64> {
+    unimplemented!()
 }
 
-/// Potential density anomaly with reference to sea pressure of 3000 dbar
-/// (75-term polynomial approximation)
-pub fn sigma3(sa: f64, ct: f64) -> Result<f64> {
-    Ok(rho(sa, ct, 3000.0)? - 1000.0)
+/// Dynamic enthalpy of seawater (75-term polynomial approximation)
+///
+/// # Arguments
+///
+/// * `sa`: Absolute Salinity \[g kg-1\]
+/// * `ct`: Conservative Temperature (ITS-90) \[deg C\]
+/// * `p`: sea pressure \[dbar\] (i.e. absolute pressure - 10.1325 dbar)
+///
+/// # Returns
+///
+/// * Dynamic enthalpy \[ J kg-1 \]
+///
+/// # References
+///
+/// * Young, W. R. (2010). Dynamic Enthalpy, Conservative Temperature, and the
+///   Seawater Boussinesq Approximation, Journal of Physical Oceanography,
+///   40(2), 394-400. doi: 10.1175/2009JPO4294.1
+///
+/// # Example:
+/// ```
+/// use gsw::volume::dynamic_enthalpy;
+/// let h_hat = dynamic_enthalpy(32.0, 10.0, 100.0).unwrap();
+/// assert!((h_hat - 975.8669302190772).abs() <= f64::EPSILON);
+/// ```
+pub fn dynamic_enthalpy(sa: f64, ct: f64, p: f64) -> Result<f64> {
+    let s: f64 = non_dimensional_sa(sa)?;
+    let tau: f64 = ct / GSW_CTU;
+    let pi: f64 = non_dimensional_p(p);
+
+    let dynamic_enthalpy_part = pi
+        * (H001
+            + s * (H101 + s * (H201 + s * (H301 + s * (H401 + s * (H501 + H601 * s)))))
+            + tau
+                * (H011
+                    + s * (H111 + s * (H211 + s * (H311 + s * (H411 + H511 * s))))
+                    + tau
+                        * (H021
+                            + s * (H121 + s * (H221 + s * (H321 + H421 * s)))
+                            + tau
+                                * (H031
+                                    + s * (H131 + s * (H231 + H331 * s))
+                                    + tau
+                                        * (H041
+                                            + s * (H141 + H241 * s)
+                                            + tau * (H051 + H151 * s + H061 * tau)))))
+            + pi * (H002
+                + s * (H102 + s * (H202 + s * (H302 + s * (H402 + H502 * s))))
+                + tau
+                    * (H012
+                        + s * (H112 + s * (H212 + s * (H312 + H412 * s)))
+                        + tau
+                            * (H022
+                                + s * (H122 + s * (H222 + H322 * s))
+                                + tau
+                                    * (H032
+                                        + s * (H132 + H232 * s)
+                                        + tau * (H042 + H142 * s + H052 * tau))))
+                + pi * (H003
+                    + s * (H103 + s * (H203 + s * (H303 + H403 * s)))
+                    + tau
+                        * (H013
+                            + s * (H113 + s * (H213 + H313 * s))
+                            + tau
+                                * (H023
+                                    + s * (H123 + H223 * s)
+                                    + tau * (H033 + H133 * s + H043 * tau)))
+                    + pi * (H004
+                        + s * (H104 + H204 * s)
+                        + tau * (H014 + H114 * s + H024 * tau)
+                        + pi * (H005 + H105 * s + H015 * tau + pi * (H006 + H007 * pi))))));
+
+    Ok(dynamic_enthalpy_part * DB2PA * 1e4)
 }
 
-/// Potential density anomaly with reference to sea pressure of 4000 dbar
-/// (75-term polynomial approximation)
-pub fn sigma4(sa: f64, ct: f64) -> Result<f64> {
-    Ok(rho(sa, ct, 4000.0)? - 1000.0)
+#[cfg(test)]
+mod test_dynamic_enthalpy {
+    use super::{dynamic_enthalpy, Error};
+
+    #[test]
+    // NaN input results in NaN output.
+    // Other libraries using GSW-rs might rely on this behavior to propagate
+    // and handle invalid elements.
+    fn nan() {
+        let h_hat = dynamic_enthalpy(f64::NAN, 1.0, 1.0);
+        assert!(h_hat.unwrap().is_nan());
+
+        let h_hat = dynamic_enthalpy(1.0, f64::NAN, 1.0);
+        assert!(h_hat.unwrap().is_nan());
+
+        let h_hat = dynamic_enthalpy(1.0, 1.0, f64::NAN);
+        assert!(h_hat.unwrap().is_nan());
+    }
+
+    #[test]
+    fn negative_sa() {
+        let h_hat = dynamic_enthalpy(-0.1, 10.0, 100.0);
+
+        if cfg!(feature = "compat") {
+            assert!((h_hat.unwrap() - 1000.0132803364188).abs() <= f64::EPSILON);
+        } else if cfg!(feature = "invalidasnan") {
+            assert!(h_hat.unwrap().is_nan());
+        } else {
+            match h_hat {
+                Err(Error::NegativeSalinity) => (),
+                _ => assert!(false),
+            }
+        }
+    }
+}
+
+/// Absolute salinity of seawater from given density, Conservative
+/// Temperature, and pressure.
+///
+/// # Notes:
+///
+/// - According to the Matlab GSW toolbox, after two iterations of this
+/// modified Newton-Raphson iteration, the error in SA is no larger than
+/// 8e-13 g/kg, which is machine precision for this calculation.
+pub fn sa_from_rho(rho: f64, ct: f64, p: f64) -> Result<f64> {
+    let v_lab = 1.0 / rho;
+    let v_0 = specvol(0.0, ct, p)?;
+    let v_50 = specvol(50.0, ct, p)?;
+
+    // First guess, a linear ratio
+    let sa = 50.0 * (v_lab - v_0) / (v_50 - v_0);
+
+    if (sa < 0.0) || (sa > 50.0) {
+        if cfg!(feature = "invalidasnan") {
+            return Ok(f64::NAN);
+        } else {
+            return Err(Error::Undefined);
+        }
+    }
+
+    // First guess of dv/dSA
+    let v_sa: f64 = (v_50 - v_0) / 50.0;
+
+    // Modified Newton-Raphson iterative optimization
+    for _ in 0..2 {
+        let sa_old = sa;
+        let delta_v = specvol(sa_old, ct, p)? - v_lab;
+        // this is half way through the modified N-R method (McDougall and
+        // Wotherspoon, 2012, appud Matlab GSW implementation)
+        let sa = sa_old - delta_v / v_sa;
+        let sa_mean = 0.5 * (sa + sa_old);
+        let (v_sa, _, _) = specvol_first_derivatives(sa_mean, ct, p)?;
+        let sa = sa_old - delta_v / v_sa;
+        if (sa < 0.0) || (sa > 50.0) {
+            if cfg!(feature = "invalidasnan") {
+                return Ok(f64::NAN);
+            } else {
+                return Err(Error::Undefined);
+            }
+        }
+    }
+    Ok(sa)
+}
+
+fn ct_from_rho() {
+    unimplemented!()
+}
+
+fn ct_maxdensity() {
+    unimplemented!()
 }
 
 #[cfg(test)]
 mod tests {
     use super::{alpha, beta, specvol, specvol_anom_standard, GSW_SSO};
 
-    #[test]
-    // Test value from Roquet 2015, Appendix C.3
-    // rounded to 9.732819628e-04
-    fn test_specvol_roquet2015() {
-        assert!((specvol(30., 10., 1000.0).unwrap() - 9.732819628e-04).abs() <= 5e-14);
-    }
-
-    #[allow(clippy::excessive_precision)]
-    #[test]
-    fn test_specvol() {
-        if cfg!(feature = "compat") {
-            // Test value from C library.
-            assert!(
-                (specvol(34.507499465692057, 27.994827331978655, 0.0).unwrap()
-                    - 0.00097855432330275953)
-                    .abs()
-                    < f64::EPSILON
-            );
-        }
-    }
-
     #[test]
     /// This anomaly should be zero for SSO & CT=0 at any depth
     fn test_specvol_anom_standard_at_standard() {
@@ -484,3 +696,99 @@ mod tests {
         }
     }
 }
+
+pub fn specvol_first_derivatives(sa: f64, ct: f64, p: f64) -> Result<(f64, f64, f64)> {
+    let xs: f64 = non_dimensional_sa(sa)?;
+    let ys: f64 = ct / GSW_CTU;
+    let z: f64 = non_dimensional_p(p);
+
+    let v_ct_part: f64 = A000
+        + xs * (A100 + xs * (A200 + xs * (A300 + xs * (A400 + A500 * xs))))
+        + ys * (A010
+            + xs * (A110 + xs * (A210 + xs * (A310 + A410 * xs)))
+            + ys * (A020
+                + xs * (A120 + xs * (A220 + A320 * xs))
+                + ys * (A030 + xs * (A130 + A230 * xs) + ys * (A040 + A140 * xs + A050 * ys))))
+        + z * (A001
+            + xs * (A101 + xs * (A201 + xs * (A301 + A401 * xs)))
+            + ys * (A011
+                + xs * (A111 + xs * (A211 + A311 * xs))
+                + ys * (A021 + xs * (A121 + A221 * xs) + ys * (A031 + A131 * xs + A041 * ys)))
+            + z * (A002
+                + xs * (A102 + xs * (A202 + A302 * xs))
+                + ys * (A012 + xs * (A112 + A212 * xs) + ys * (A022 + A122 * xs + A032 * ys))
+                + z * (A003 + A103 * xs + A013 * ys + A004 * z)));
+
+    let v_ct = 0.025 * v_ct_part;
+
+    let v_sa_part: f64 = B000
+        + xs * (B100 + xs * (B200 + xs * (B300 + xs * (B400 + B500 * xs))))
+        + ys * (B010
+            + xs * (B110 + xs * (B210 + xs * (B310 + B410 * xs)))
+            + ys * (B020
+                + xs * (B120 + xs * (B220 + B320 * xs))
+                + ys * (B030 + xs * (B130 + B230 * xs) + ys * (B040 + B140 * xs + B050 * ys))))
+        + z * (B001
+            + xs * (B101 + xs * (B201 + xs * (B301 + B401 * xs)))
+            + ys * (B011
+                + xs * (B111 + xs * (B211 + B311 * xs))
+                + ys * (B021 + xs * (B121 + B221 * xs) + ys * (B031 + B131 * xs + B041 * ys)))
+            + z * (B002
+                + xs * (B102 + xs * (B202 + B302 * xs))
+                + ys * (B012 + xs * (B112 + B212 * xs) + ys * (B022 + B122 * xs + B032 * ys))
+                + z * (B003 + B103 * xs + B013 * ys + B004 * z)));
+
+    let v_sa = 0.5 * GSW_SFAC * v_sa_part / xs;
+
+    let v_p_part = C000
+        + xs * (C100 + xs * (C200 + xs * (C300 + xs * (C400 + C500 * xs))))
+        + ys * (C010
+            + xs * (C110 + xs * (C210 + xs * (C310 + C410 * xs)))
+            + ys * (C020
+                + xs * (C120 + xs * (C220 + C320 * xs))
+                + ys * (C030 + xs * (C130 + C230 * xs) + ys * (C040 + C140 * xs + C050 * ys))))
+        + z * (C001
+            + xs * (C101 + xs * (C201 + xs * (C301 + C401 * xs)))
+            + ys * (C011
+                + xs * (C111 + xs * (C211 + C311 * xs))
+                + ys * (C021 + xs * (C121 + C221 * xs) + ys * (C031 + C131 * xs + C041 * ys)))
+            + z * (C002
+                + xs * (C102 + C202 * xs)
+                + ys * (C012 + C112 * xs + C022 * ys)
+                + z * (C003 + C103 * xs + C013 * ys + z * (C004 + C005 * z))));
+
+    let v_p = 1e-8 * v_p_part;
+
+    Ok((v_sa, v_ct, v_p))
+}
+
+/*
+fn specvol_first_derivatives_wrt_enthalpy(sa: f64, ct: f64, p: f64) -> (f64, f64) {
+    unimplemented!()
+}
+
+fn specvol_ice(t: f64, p: f64) {
+    unimplemented!()
+}
+
+fn specvol_second_derivatives(sa: f64, ct: f64, p: f64) -> (f64, f64, f64, f64, f64) {
+    unimplemented!()
+}
+
+fn specvol_second_derivatives_wrt_enthalpy(sa: f64, ct: f64, p: f64) -> (f64, f64, f64) {
+    unimplemented!()
+}
+
+fn specvol_diff() {
+    unimplemented!()
+}
+fn specvol_from_pot_enthalpy_ice() {
+    unimplemented!()
+}
+fn specvol_from_pot_enthalpy_ice_poly() {
+    unimplemented!()
+}
+fn specvol_p_parts() {
+    unimplemented!()
+}
+*/
